#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#ifndef UNICODE
#define UNICODE
#endif
#ifndef _UNICODE
#define _UNICODE
#endif

#include <stdlib.h>

#include <winstrct.h>
#include <shellapi.h>
#include <wconsole.h>
#include <wfind.h>

#if defined(_DLL) && defined(_M_IX86)
#pragma comment(lib, "winstrct.lib")
#endif

#pragma comment(lib, "advapi32.lib")
#pragma comment(lib, "shell32.lib")

#ifndef FILE_ATTRIBUTE_RECALL_ON_OPEN
#define FILE_ATTRIBUTE_RECALL_ON_OPEN                    0x00040000  
#endif
#ifndef FILE_ATTRIBUTE_RECALL_ON_DATA_ACCESS
#define FILE_ATTRIBUTE_RECALL_ON_DATA_ACCESS             0x00400000  
#endif
#define FILE_OFFLINE_ATTRIBUTES (FILE_ATTRIBUTE_OFFLINE | FILE_ATTRIBUTE_RECALL_ON_OPEN | FILE_ATTRIBUTE_RECALL_ON_DATA_ACCESS | FILE_ATTRIBUTE_VIRTUAL)

bool follow_reparse_points = true;
bool include_offline = false;
bool count_files = false;

enum PRESENTATION_FORMAT
{
    PRESENTATION_FORMAT_BYTES,
    PRESENTATION_FORMAT_BLOCKS,
    PRESENTATION_FORMAT_HUMAN_READABLE,
    PRESENTATION_FORMAT_KB,
    PRESENTATION_FORMAT_MB,
    PRESENTATION_FORMAT_GB,
    PRESENTATION_FORMAT_TB
} presentation_format;

//
// These classes are used to track multiple hard links to files.
//

// This needs to be a power of 2
#define FILE_HASH_TABLE_SIZE 1024

class FileTableClass
{
    ULARGE_INTEGER FileNumber;
    DWORD VolumeSerialNumber;
    FileTableClass *Next;

    FileTableClass(ULARGE_INTEGER &NewFileNumber,
        DWORD NewVolumeSerialNumber,
        FileTableClass *NewNext) :
        FileNumber(NewFileNumber),
        VolumeSerialNumber(NewVolumeSerialNumber),
        Next(NewNext)
    {
    }

public:

    // This function adds a file to the list if it is not already in the list.
    // Returns true if the list already held the file, false otherwise.
    static bool FindOrAdd(DWORD dwLowPart, DWORD dwHighPart,
        DWORD dwVolumeSerialNumber)
    {
        ULARGE_INTEGER FileNumber;
        FileNumber.HighPart = dwHighPart;
        FileNumber.LowPart = dwLowPart;

        for (FileTableClass *Ptr =
            FileHashTable[FileNumber.QuadPart & (FILE_HASH_TABLE_SIZE - 1)];
            Ptr != NULL;
        Ptr = Ptr->Next)
            if ((Ptr->FileNumber.QuadPart == FileNumber.QuadPart) &
                (Ptr->VolumeSerialNumber == dwVolumeSerialNumber))
                return true;

        FileHashTable[FileNumber.QuadPart & (FILE_HASH_TABLE_SIZE - 1)] =
            new FileTableClass(FileNumber, dwVolumeSerialNumber,
            FileHashTable[FileNumber.QuadPart &
            (FILE_HASH_TABLE_SIZE - 1)]);

        return false;
    }
} *FileHashTable[FILE_HASH_TABLE_SIZE] = { NULL };

LONGLONG
directory(LPWSTR file, SIZE_T filechars, LPWSTR fptr, DWORD dwLastClusterMask);

void print_file_info(LONGLONG TotalFileSize, LPCSTR oem_file);

int
wmain(int argc, LPWSTR *argv)
{
    bool displayhelp = false;

    presentation_format = PRESENTATION_FORMAT_BYTES;

    SetOemPrintFLineLength(GetStdHandle(STD_ERROR_HANDLE));

    if (argc - 1)
        while ((argv[1][0] | 0x02) == L'/')
        {
            while (*(++argv[1]))
                switch (*argv[1] | 0x20)
            {
                case L'x':
                    follow_reparse_points = false;
                    continue;
                case L'o':
                    include_offline = true;
                    continue;
                case L'b':
                    presentation_format = PRESENTATION_FORMAT_BLOCKS;
                    continue;
                case L'h':
                    presentation_format = PRESENTATION_FORMAT_HUMAN_READABLE;
                    continue;
                case L'k':
                    presentation_format = PRESENTATION_FORMAT_KB;
                    continue;
                case L'm':
                    presentation_format = PRESENTATION_FORMAT_MB;
                    continue;
                case L'g':
                    presentation_format = PRESENTATION_FORMAT_GB;
                    continue;
                case L't':
                    presentation_format = PRESENTATION_FORMAT_TB;
                    continue;
                case L'c':
                    count_files = true;
                    continue;
                default:
                    displayhelp = true;
                    continue;
            }
            ++argv;
            --argc;

            if (argc == 1)
                break;
        }

    if (argc == 1)
        displayhelp = true;

    if (displayhelp)
    {
        puts("sizeof, freeware by Olof Lagerkvist.\r\n"
            "http://www.ltr-data.se   e-mail: olof@ltr-data.se\r\n"
            "More info, including distribution permissions and\r\n"
            "source code is available on the website.\r\n"
            "\n"
            "This tool displays the total allocation size on disk for files and directories.\r\n"
            "\n"
            "Command line syntax:\r\n"
            "sizeof [-xo] [-presentation] file1 [file2 ...]\r\n"
            "\n"
            "-x       Do not follow junctions or other reparse points. This makes sure that\r\n"
            "         sizeof stays within the filesystem where it starts searching.\r\n"
            "\n"
            "-o       Include sizes of offline files.\r\n"
            "\n"
            "-c       Count and display number of files in each found directory, instead of\r\n"
            "         count and display sizes of them.\r\n"
            "\n"
            "The presentation determines in which format to display the sizes of files and\r\n"
            "directories. The argument can be any of the following:\r\n\n"
            "-b       Number of 512-byte blocks.\r\n"
            "-h       Human readable format.\r\n"
            "-k       Kilobytes.\r\n"
            "-m       Megabytes.\r\n"
            "-g       Gigabytes.\r\n"
            "-t       Terabytes.");
        return 0;
    }

    EnableBackupPrivileges();

    WHeapMem<WCHAR> file(UNICODE_STRING_MAX_BYTES, HEAP_GENERATE_EXCEPTIONS);

    LONGLONG summary_size = 0;

    while (argc-- > 1)
    {
        ++argv;

        if (wcslen(argv[0]) > file.Count())
        {
            oem_printf(stderr, "Too long path: %1!ws!%%n", argv[0]);
            continue;
        }

        wcscpy(file, argv[0]);

        LPWSTR filepart = wcsrchr(file, L'/');
        if (filepart != NULL)
            *filepart = L'\\';

        filepart = wcsrchr(file, L'\\');
        if (filepart != NULL)
            ++filepart;
        else
            filepart = file;

        WFileFinder finder(file);
        if (!finder)
        {
            win_perror(file);
            continue;
        }

        LPCWSTR specialdirnode = NULL;
        if (wcscmp(filepart, L".") == 0)
            specialdirnode = L".";
        else if (wcscmp(filepart, L"..") == 0)
            specialdirnode = L"..";

        *filepart = 0;

        DWORD dwBytesPerSector;
        DWORD dwSectorsPerCluster;
        DWORD dwFreeClusters;
        DWORD dwTotalClusters;
        if ((!GetDiskFreeSpace(file, &dwSectorsPerCluster, &dwBytesPerSector,
            &dwFreeClusters, &dwTotalClusters)) &&
            (!GetDiskFreeSpace(NULL, &dwSectorsPerCluster, &dwBytesPerSector,
                &dwFreeClusters, &dwTotalClusters)))
        {
            win_perror();
            return -1;
        }

        DWORD BytesPerCluster = dwSectorsPerCluster * dwBytesPerSector;
        DWORD dwClusterMask = 0;
        while (BytesPerCluster >>= 1)
            ++dwClusterMask;

        do
        {
            LARGE_INTEGER TotalFileSize;

            if ((wcscmp(finder.cFileName, L".") == 0) ||
                wcscmp(finder.cFileName, L"..") == 0)
            {
                continue;
            }

            if (specialdirnode != NULL)
                wcscpy(finder.cFileName, specialdirnode);

            *filepart = 0;
            if (wcslen(file) + wcslen(finder.cFileName) >= file.Count())
            {
                oem_printf(stderr,
                    "Too long path: %1!ws!%2!ws!%%n", file, finder.cFileName);

                if (finder.Next())
                    continue;
                else
                    break;
            }

            wcscpy(filepart, finder.cFileName);

            WHeapMem<char> oem_file_mem(WideToOemAlloc(file));
            LPCSTR oem_file = oem_file_mem;

            if ((!follow_reparse_points) &&
                (finder.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) != 0)
            {
                continue;
            }
            else if ((!include_offline) &&
                (finder.dwFileAttributes & FILE_OFFLINE_ATTRIBUTES) != 0)
            {
                continue;
            }
            else if ((finder.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
            {
                if (wcslen(file) + 13 >= file.Count())
                {
                    oem_printf(stderr,
                        "Too long path: %1!ws!%2!ws!%%n", file, finder.cFileName);

                    if (finder.Next())
                        continue;
                    else
                        break;
                }

                wcscat(filepart, L"\\");

                TotalFileSize.QuadPart = directory(file, file.Count(), filepart + wcslen(filepart), dwClusterMask);
            }
            // If just counting number of files
            else if (count_files)
            {
                TotalFileSize.QuadPart = 1;
            }
            // If file is compressed, get compressed size before calculations
            else if ((finder.dwFileAttributes &
                (FILE_ATTRIBUTE_COMPRESSED | FILE_ATTRIBUTE_SPARSE_FILE)) != 0)
            {
                TotalFileSize.LowPart =
                    GetCompressedFileSize(file, (LPDWORD)&TotalFileSize.HighPart);

                if (TotalFileSize.LowPart == 0xFFFFFFFF && GetLastError() != NO_ERROR)
                {
                    win_perror(file);
                    continue;
                }
            }
            else
            {
                TotalFileSize.HighPart = finder.nFileSizeHigh;
                TotalFileSize.LowPart = finder.nFileSizeLow;

                TotalFileSize.QuadPart =
                    (((TotalFileSize.QuadPart - 1) >> dwClusterMask) + 1) <<
                    dwClusterMask;
            }

            print_file_info(TotalFileSize.QuadPart, oem_file);

            summary_size += TotalFileSize.QuadPart;
        } while (finder.Next());
    }

    print_file_info(summary_size, "(Total)");
}

void print_file_info(LONGLONG TotalFileSize, LPCSTR oem_file)
{
    switch (presentation_format)
    {
    case PRESENTATION_FORMAT_BLOCKS:
        printf("%I64i\t%s\n", TotalFileSize >> 9, oem_file);
        break;
    case PRESENTATION_FORMAT_HUMAN_READABLE:
        printf("%.4g %s\t%s\n", TO_h(TotalFileSize), TO_p(TotalFileSize), oem_file);
        break;
    case PRESENTATION_FORMAT_KB:
        printf("%I64i\t%s\n", TotalFileSize >> 10, oem_file);
        break;
    case PRESENTATION_FORMAT_MB:
        printf("%I64i\t%s\n", TotalFileSize >> 20, oem_file);
        break;
    case PRESENTATION_FORMAT_GB:
        printf("%I64i\t%s\n", TotalFileSize >> 30, oem_file);
        break;
    case PRESENTATION_FORMAT_TB:
        printf("%I64i\t%s\n", TotalFileSize >> 40, oem_file);
        break;
    default:
        printf("%I64i\t%s\n", TotalFileSize, oem_file);
        break;
    }
}

LONGLONG
directory(LPWSTR file, SIZE_T filechars, LPWSTR fptr, DWORD dwLastClusterMask)
{
    LONGLONG TotalFileSize = 0;
	DWORD dwBytesPerSector;
	DWORD dwSectorsPerCluster;
	DWORD dwFreeClusters;
	DWORD dwTotalClusters;
    DWORD dwClusterMask = 0;

    if (GetDiskFreeSpace(file, &dwSectorsPerCluster, &dwBytesPerSector,
        &dwFreeClusters, &dwTotalClusters))
    {
        DWORD BytesPerCluster = dwSectorsPerCluster * dwBytesPerSector;
        dwClusterMask = 0;
        while (BytesPerCluster >>= 1)
            ++dwClusterMask;
    }
    else
    {
        dwClusterMask = dwLastClusterMask;
    }

    fptr[0] = L'*';
    fptr[1] = 0;

    Sleep(0);			// Let others run

    WFileFinder finddata(file);

    if (finddata)
    {
        do
        {
            if (((finddata.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) != 0) &&
                !follow_reparse_points)
            {
                continue;
            }

            if (finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
            {
                if (!(wcscmp(finddata.cFileName, L".") == 0) &&
                    !(wcscmp(finddata.cFileName, L"..") == 0))
                {
                    if (wcslen(finddata.cFileName) + 13 + (SIZE_T)(fptr - file) >= filechars)
                    {
                        *fptr = 0;
                        oem_printf(stderr,
                            "Too long path: %1!ws!%2!ws!%%n",
                            file, finddata.cFileName);
                        continue;
                    }

                    wcscpy(fptr, finddata.cFileName);
                    wcscat(fptr, L"\\");

                    TotalFileSize += directory(file, filechars, fptr + wcslen(fptr), dwClusterMask);
                }
            }
            else if (count_files)
            {
                ++TotalFileSize;
            }
            else
            {
                if (wcslen(finddata.cFileName) + (SIZE_T)(fptr - file) >= filechars)
                {
                    *fptr = 0;
                    oem_printf(stderr,
                        "Too long path: %1!ws!%2!ws!%%n", file, finddata.cFileName);

                    continue;
                }

                wcscpy(fptr, finddata.cFileName);

                bool skip_this = false;

                if ((!include_offline) && (finddata.dwFileAttributes &
                    FILE_OFFLINE_ATTRIBUTES))
                {
                    skip_this = true;
                }
                else
                {
                    HANDLE hFile =
                        CreateFile(file, 0, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                            OPEN_EXISTING, 0, NULL);

                    if (hFile != INVALID_HANDLE_VALUE)
                    {
                        BY_HANDLE_FILE_INFORMATION file_info;
                        if (GetFileInformationByHandle(hFile, &file_info))
                        {
                            skip_this =
                                FileTableClass::
                                FindOrAdd(file_info.nFileIndexLow,
                                    file_info.nFileIndexHigh & 0xFFFF,
                                    file_info.dwVolumeSerialNumber);
                        }

                        CloseHandle(hFile);
                    }
                }

                LARGE_INTEGER filesize;

                // If file is compressed, get compressed size before calculations
                if (skip_this)
                {
                }
                else if (finddata.dwFileAttributes &
                    (FILE_ATTRIBUTE_COMPRESSED | FILE_ATTRIBUTE_SPARSE_FILE))
                {
                    filesize.LowPart =
                        GetCompressedFileSize(file, (LPDWORD)&filesize.HighPart);

                    if ((filesize.LowPart == 0xFFFFFFFF) ?
                        (GetLastError() != NO_ERROR) : false)
                        win_perror(file);
                    else
                        TotalFileSize += filesize.QuadPart;
                }
                else
                {
                    filesize.LowPart = finddata.nFileSizeLow;
                    filesize.HighPart = finddata.nFileSizeHigh;

                    TotalFileSize +=
                        (((filesize.QuadPart - 1) >> dwClusterMask) + 1) <<
                        dwClusterMask;
                }

            }
        }
        while (finddata.Next());
    }

    return TotalFileSize;
}

// This is enough startup code for this program if compiled to use the DLL CRT.
#if !defined(_DEBUG) && !defined(DEBUG) && _MSC_VER < 1900
extern "C" int
wmainCRTStartup()
{
    int argc = 0;
    LPWSTR *argv = CommandLineToArgvW(GetCommandLine(), &argc);
    if (argv == NULL)
    {
#ifndef _M_ARM
        MessageBoxA(NULL, "This program requires Windows NT.",
            "Incompatible Windows version", MB_ICONERROR);
#endif
        ExitProcess((UINT)-1);
    }

    exit(wmain(argc, argv));
}
#endif
